/*
    CFS0.11 filesystem implementation for FUSE
    Copyright (C) 2005-2020 Kajtar Zsolt (Soci/Singular) <soci@c64.rulez.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/
#include "petscii.h"
#include <wchar.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

int petscii_from_ascii(Petscii *s2, const char *s, ssize_t len, size_t max) {
    mbstate_t ps;
    memset(&ps, 0, sizeof(mbstate_t));
    wchar_t w;
    ssize_t l;
    while (*s && len > 0 && max) {
        if (*s == '\\' && (s[1] == 'u' || s[1] == 'U')) {
            int j;
            w = 0; l = 2;
            for (j = (s[1] == 'u') ? 12 : 28; j >= 0; j -= 4, l++) {
                if (s[l] >= '0' && s[l] <= '9') w |= (s[l] - '0') << j;
                else if (s[l] >= 'a' && s[l] <= 'f') w |= (s[l] - 'a' + 10) << j;
                else {
                    w = *s;
                    l = 1;
                    break;
                }
            }
        } else if (*s >= 0x20 && *s <= 0x7e) {
            w = *s;
            l = 1;
        } else {
            l = mbrtowc(&w, s, MB_CUR_MAX, &ps);
        }
        if (l < 0) return 1;
        s += l; len -= l;
        if (w >= 0x20 && w <= 0x40) *s2 = w;
        else if (w >= 0x41 && w <= 0x5a) *s2 = w + 0x80;
        else if (w >= 0x61 && w <= 0x7a) *s2 = w - 0x20;
        else if (w >= 0xe100 && w <= 0xe1ff) *s2 = w - 0xe100;
        else switch (w) {
        case 0x2215: *s2 = 0x2f; break;
        case 0x5b:   *s2 = 0x5b; break;
        case 0xa3:   *s2 = 0x5c; break;
        case 0x5d:   *s2 = 0x5d; break;
        case 0x2191: *s2 = 0x5e; break;
        case 0x2190: *s2 = 0x5f; break;
        case 0xa0:   *s2 = 0xa0; break;
        case 0x258c: *s2 = 0xa1; break;
        case 0x2584: *s2 = 0xa2; break;
        case 0x2594: *s2 = 0xa3; break;
        case 0x5f:   *s2 = 0xa4; break;
        case 0x258f: *s2 = 0xa5; break;
        case 0x2592: *s2 = 0xa6; break;
        case 0x2595: *s2 = 0xa7; break;
        case 0x1fb8f:*s2 = 0xa8; break;
        case 0x1fb99: *s2 = 0xa9; break;
        case 0x1fb87:*s2 = 0xaa; break;
        case 0x251c: *s2 = 0xab; break;
        case 0x2597: *s2 = 0xac; break;
        case 0x2514: *s2 = 0xad; break;
        case 0x2510: *s2 = 0xae; break;
        case 0x2582: *s2 = 0xaf; break;
        case 0x250c: *s2 = 0xb0; break;
        case 0x2534: *s2 = 0xb1; break;
        case 0x252c: *s2 = 0xb2; break;
        case 0x2524: *s2 = 0xb3; break;
        case 0x258e: *s2 = 0xb4; break;
        case 0x258d: *s2 = 0xb5; break;
        case 0x1fb88:*s2 = 0xb6; break;
        case 0x1fb82:*s2 = 0xb7; break;
        case 0x1fb83:*s2 = 0xb8; break;
        case 0x2583: *s2 = 0xb9; break;
        case 0x2713: *s2 = 0xba; break;
        case 0x2596: *s2 = 0xbb; break;
        case 0x259d: *s2 = 0xbc; break;
        case 0x2518: *s2 = 0xbd; break;
        case 0x2598: *s2 = 0xbe; break;
        case 0x259a: *s2 = 0xbf; break;
        case 0x2500: *s2 = 0xc0; break;
        case 0x253c: *s2 = 0xdb; break;
        case 0x1fb8c:*s2 = 0xdc; break;
        case 0x2502: *s2 = 0xdd; break;
        case 0x1fb98:*s2 = 0xdf; break;
        case 0x03c0: *s2 = 0xff; break;
        default: return 1;
        }
        s2++; max--;
    }
    *s2 = 0;
    return 0;
}

size_t petscii_to_ascii(char *st2, const Petscii *s, int link) {
    wchar_t s2;
    mbstate_t ps;
    size_t i = 0, l;
    memset(&ps, 0, sizeof(mbstate_t));
    while (*s) {
        char mb[MB_CUR_MAX > 11 ? MB_CUR_MAX : 11];
        if (*s >= 0x20 && *s <= 0x40 && (*s != 0x2f || link)) s2 = *s;
        else if (*s >= 0x41 && *s <= 0x5a) s2 = *s + 0x20;
        else if (*s >= 0xc1 && *s <= 0xda) s2 = *s - 0x80;
        else switch (*s) {
        case 0x2f: s2 = 0x2215; break;
        case 0x5b: s2 = 0x5b; break;
        case 0x5c: s2 = 0xa3; break;
        case 0x5d: s2 = 0x5d; break;
        case 0x5e: s2 = 0x2191; break;
        case 0x5f: s2 = 0x2190; break;
        case 0xa0: s2 = 0xa0; break;
        case 0xa1: s2 = 0x258c; break;
        case 0xa2: s2 = 0x2584; break;
        case 0xa3: s2 = 0x2594; break;
        case 0xa4: s2 = 0x5f; break;
        case 0xa5: s2 = 0x258f; break;
        case 0xa6: s2 = 0x2592; break;
        case 0xa7: s2 = 0x2595; break;
        case 0xa8: s2 = 0x1fb8f; break;
        case 0xa9: s2 = 0x1fb99; break;
        case 0xaa: s2 = 0x1fb87; break;
        case 0xab: s2 = 0x251c; break;
        case 0xac: s2 = 0x2597; break;
        case 0xad: s2 = 0x2514; break;
        case 0xae: s2 = 0x2510; break;
        case 0xaf: s2 = 0x2582; break;
        case 0xb0: s2 = 0x250c; break;
        case 0xb1: s2 = 0x2534; break;
        case 0xb2: s2 = 0x252c; break;
        case 0xb3: s2 = 0x2524; break;
        case 0xb4: s2 = 0x258e; break;
        case 0xb5: s2 = 0x258d; break;
        case 0xb6: s2 = 0x1fb88; break;
        case 0xb7: s2 = 0x1fb82; break;
        case 0xb8: s2 = 0x1fb83; break;
        case 0xb9: s2 = 0x2583; break;
        case 0xba: s2 = 0x2713; break;
        case 0xbb: s2 = 0x2596; break;
        case 0xbc: s2 = 0x259d; break;
        case 0xbd: s2 = 0x2518; break;
        case 0xbe: s2 = 0x2598; break;
        case 0xbf: s2 = 0x259a; break;
        case 0xc0: s2 = 0x2500; break;
        case 0xdb: s2 = 0x253c; break;
        case 0xdc: s2 = 0x1fb8c; break;
        case 0xdd: s2 = 0x2502; break;
        case 0xdf: s2 = 0x1fb98; break;
        case 0xff: s2 = 0x03c0; break;
        default: s2 = 0xe100 + *s;
        }
        s++;
        if (s2 >= 0x20 && s2 <= 0x7e) {
            st2[i++] = s2;
            continue;
        }
        l = wcrtomb(mb, s2, &ps);
        if (l == (size_t)-1) {
            l = sprintf(mb, s2 <= 0xffff ? "\\u%04x" : "\\U%08x", (unsigned int)s2);
        }
        memcpy(st2 + i, mb, l);
        i += l;
    }
    st2[i] = 0;
    return i;
}

